
/* C I/O functions */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#include "global.h"
#include "error.h"
#include "mem.h"
#include "codegen.h"
#include "optimize.h"
#include "io.h"



/*****************************************************************************/
/*	  			     data				     */
/*****************************************************************************/



/* Input line stuff */
char linebuf [LINESIZE];
char* line = linebuf;
char* lptr = 0;

/* The output buffer */
static unsigned outfill = 0;
char outq [8192];
char* outqi = outq;

/* Input file table and number of open input files */
struct filent filetab[MAXFILES];
int ifile = 0;

/* Current input file stream data */
FILE* inp = 0;
char* fin = 0;
unsigned ln = 0;



/*****************************************************************************/
/*			    output buffer handling			     */
/*****************************************************************************/



void flushout (void)
/* Flush output queue */
{
    outfill = OptAddBuffer (outq, outqi - outq);
    outqi = outq + outfill;
}



void outchar (char c)
/* Output char c to assembler file.  Really just goes out to
 * buffer, so the optimizer can work on it.
 */
{
    if (outfill >= sizeof (outq)) {
     	/* Forced empty buffer. This will in some cases mean, that we get
     	 * garbled output since other functions may hold pointers into the
     	 * data to kill stuff. Print a warning if that happens.
     	 */
       	Warning (WARN_OUTPUT_FLUSH);
     	flushout ();
    }
    *outqi++ = c;
    ++outfill;
}



void outstr (char* s)
/* Output string s to assembler file.  Really just goes out to
 * buffer, so the optimizer can work on it.
 */
{
    char c;
    while ((c = *s)) {
	outchar (c);
	++s;
    }
}



void outnl (void)
/* Output a newline into the output buffer */
{
    outchar ('\n');
}



void outline (char* s)
/* Output the given string to the assembler file and add a newline */
{
    outstr (s);
    outnl ();
}



void out (char* format, ...)
/* Output a variable string to the assembler file */
{
    va_list ap;
    char buf [1024];

    va_start (ap, format);
    vsprintf (buf, format, ap);
    va_end (ap);

    outstr (buf);
}



void outnum16 (unsigned num)
/* Output a 16bit decimal number to the assembler file */
{
    out ("$%04X", num & 0xFFFF);
}



void outnum32 (unsigned long num)
/* Output a 32bit decimal number to the assembler file */
{
    out ("$%08X", num);
}



/*****************************************************************************/
/*  	       	   		     code				     */
/*****************************************************************************/



int nch (void)
/* Get the next char in input stream (the one behind the current one) */
{
    if (*lptr == '\0') {
    	return 0;
    } else {
    	return lptr[1] & 0xFF;
    }
}



int cgch (void)
/* Get the current character in the input stream and advance line
 * pointer (unless already at end of line).
 */
{
    if (*lptr == '\0') {
   	return (0);
    } else {
   	return (*lptr++ & 0xFF);
    }
}



int gch (void)
/* Get the current character in the input stream and advance line
 * pointer (no end of line check is performed).
 */
{
    return (*lptr++ & 0xFF);
}



void kill (void)
/* Reset input line pointer, clear input line */
{
    lptr = line;
    *lptr = '\0';
}



int readline (void)
/* Get a line from the current input.  Returns -1 on end of file. */
{
    int len;
    struct filent *pftab;

    /* Read lines until we get one with real contents */
    do {

    	kill ();
    	if (inp == 0) {
    	    return 0;
    	}
       	if (fgets (line, LINESIZE, inp) == 0) {
    	    /* eof */
    	    len = 0;
    	    *line = '\0';
    	    fclose(inp);

    	    /* Leave include file */
    	    if (ifile > 0) {
    		xfree (fin);
    		inp = (pftab = &filetab[--ifile])->f_iocb;
    		ln = pftab->f_ln;
    		fin = pftab->f_name;
    	    } else {
    		inp = 0;
    	    }
    	} else {
	    /* We got a new line */
    	    ++ln;

	    /* Remove the trailing newline if we have one */
    	    len = strlen (line);
	    while (len > 0 && line [len-1] == '\n') {
		--len;
	    }
	    line [len] = '\0';
		      
	    /* Output the source line in the generated assembler file
	     * if requested.
	     */
    	    if (IncSource && len > 0) {
    		outchar (';');
    		outline (line);
    	    }
    	}

    } while (len == 0);

    /* Get a line */
    lptr = line;
    return 1;
}



